
import os, json

from usgs import api
from usgs import CWIC_LSI_EXPLORER_CATALOG_NODE
from usgs import EARTH_EXPLORER_CATALOG_NODE
from usgs import HDDS_EXPLORER_CATALOG_NODE
from usgs import LPCS_EXPLORER_CATALOG_NODE

scriptdir = os.path.dirname(os.path.abspath(__file__))


def get_datasets_in_nodes():
    """
    Get the node associated with each dataset. Some datasets
    will have an ambiguous node since they exists in more than
    one node.
    """

    data_dir = os.path.join(scriptdir, "..", "usgs", "data")

    cwic = map(lambda d: d["datasetName"], api.datasets(None, CWIC_LSI_EXPLORER_CATALOG_NODE)['data'])
    ee = map(lambda d: d["datasetName"], api.datasets(None, EARTH_EXPLORER_CATALOG_NODE)['data'])
    hdds = map(lambda d: d["datasetName"], api.datasets(None, HDDS_EXPLORER_CATALOG_NODE)['data'])
    lpcs = map(lambda d: d["datasetName"], api.datasets(None, LPCS_EXPLORER_CATALOG_NODE)['data'])

    # Create mapping from dataset to node
    datasets = {}
    datasets.update( { ds : "CWIC" for ds in cwic } )
    datasets.update( { ds : "EE" for ds in ee } )
    datasets.update( { ds : "HDDS" for ds in hdds } )
    datasets.update( { ds : "LPCS" for ds in lpcs } )

    datasets_path = os.path.join(data_dir, "datasets.json")
    with open(datasets_path, "w") as f:
        f.write(json.dumps(datasets))

    # Find the datasets with ambiguous nodes
    cwic_ee = [ds for ds in cwic if ds in ee]
    cwic_hdds = [ds for ds in cwic if ds in hdds]
    cwic_lpcs = [ds for ds in cwic if ds in lpcs]
    ee_hdds = [ds for ds in ee if ds in hdds]
    ee_lpcs = [ds for ds in ee if ds in lpcs]
    hdds_lpcs = [ds for ds in hdds if ds in lpcs]



if __name__ == '__main__':
    get_datasets_in_nodes()


import json
import math
from collections import defaultdict
from usgs import CATALOG_NODES, USGSDependencyRequired


def dataset_filters(dataset):
    """
    This request is used to return the metadata filter fields for the specified
    dataset. These values can be used as additional criteria when submitting search
    and hit queries.

    :param str dataset:
    """
    return json.dumps({
        "datasetName": dataset,
    })

def download_options(dataset, entity_ids):
    """
    The download options request is used to discover downloadable products for
    each dataset. If a download is marked as not available, an order must be
    placed to generate that product.

    :param str dataset:
    :param str entity_ids:
    """

    payload = {
        "datasetName": dataset,
        "entityIds": entity_ids
    }

    return json.dumps(payload)

def dataset_download_options(dataset):
    """
    The dataset download options request is used to discover downloadable
    products for a specified dataset. Unlike the `download_options` request,
    this does not check product availability.

    :param str dataset: Used to identify the which dataset to return results for.
    """
    payload = {"datasetName": dataset}

    return json.dumps(payload)

def download_request(dataset, entity_id, product_id):
    """
    The use of this request will be to obtain valid data download URLs.

    :param str dataset:
    :param str entity_id:
    :param str product_id:
    """

    payload = {
        "downloads": [
            {
                "entityId": entity_id,
                "productId": product_id
            }
        ],
        "downloadApplication": "EE"
    }

    return json.dumps(payload)

def dataset_search(dataset, catalog, start_date=None, end_date=None, ll=None, ur=None):
    """
    This method is used to find datasets available for searching. By passing only
    an API Key, all available datasets are returned. Additional parameters such
    as temporal range and spatial bounding box can be used to find datasets that
    provide more specific data. The dataset name parameter can be used to limit
    the results based on matching the supplied value against the public dataset
    name with assumed wildcards at the beginning and end.

    :param str dataset:
    :param str catalog:
    :param start_date:
        Used for searching scene acquisition - will accept anything
        that the PHP strtotime function can understand

    :param end_date:
        Used for searching scene acquisition - will accept anything
        that the PHP strtotime function can understand

    :param ll:
        Lower left corner of an AOI bounding box - in decimal form
        Longitude/Latitude dictionary

        e.g. { "longitude": 0.0, "latitude": 0.0 }

    :param ur:
        Upper right corner of an AOI bounding box - in decimal form
        Longitude/Latitude dictionary

        e.g. { "longitude": 0.0, "latitude": 0.0 }
    """

    payload = {
        "datasetName": dataset,
        "catalog": catalog
    }

    if start_date and end_date:
        payload["temporalFilter"] = {
            "start": start_date,
            "end": end_date
        }

    if ll and ur:
        payload["spatialFilter"] = {
            "filterType": "mbr",
            "lowerLeft": {
                "latitude": ll["latitude"],
                "longitude": ll["longitude"]
            },
            "upperRight": {
                "latitude": ur["latitude"],
                "longitude": ur["longitude"]
            }
        }

    return json.dumps(payload)

def login(username, token):
    """
    Upon a successful login, an API key will be returned. This key will be active
    for two hours and should be destroyed upon final use of the service by calling
    the logout method.

    :param str username:
    :param str token:
    """
    payload = {
        "username": username,
        "token": token
    }

    return json.dumps(payload)


def scene_metadata(dataset, entity_id):
    """
    The use of the metadata request is intended for those who have
    acquired scene IDs from a different source. It will return the
    same metadata that is available via the search request.

    :param dataset:
    :param entity_id:
    """
    payload = {
        "datasetName": dataset,
        "entityId": entity_id,
        "metadataType": "full"
    }

    return json.dumps(payload)


def great_circle_dist(lat, lng, dist):
    lat = math.radians(lat)
    lng = math.radians(lng)
    brng = math.radians(45.0)
    ibrng = math.radians(225.0)

    earth_radius = 6371000.0
    dR = (dist / 2.0)/ earth_radius

    lat1 = math.asin( math.sin(lat)*math.cos(dR) +
        math.cos(lat)*math.sin(dR)*math.cos(brng) );
    lng1 = lng + math.atan2(math.sin(brng)*math.sin(dR)*math.cos(lat),
        math.cos(dR)-math.sin(lat)*math.sin(lat1));

    lat2 = math.asin( math.sin(lat)*math.cos(dR) +
        math.cos(lat)*math.sin(dR)*math.cos(ibrng) );
    lng2 = lng + math.atan2(math.sin(ibrng)*math.sin(dR)*math.cos(lat),
        math.cos(dR)-math.sin(lat)*math.sin(lat2));

    return [math.degrees(lat1), math.degrees(lat2)], [math.degrees(lng1), math.degrees(lng2)]

def scene_search(
    dataset, max_results=None, metadata_type=None, start_date=None,
    end_date=None, ll=None, ur=None,
    lat=None, lng=None, distance=100,
    where=None, starting_number=None, scene_filter=None):

    payload = defaultdict(dict, {
        "datasetName": dataset,
        "maxResults": max_results,
        "startingNumber": starting_number,
        "metadataType": metadata_type
    })

    if (start_date is not None) and (end_date is not None):
        payload["sceneFilter"]["acquisitionFilter"] = {
            "start": start_date,
            "end": end_date
        }

    # Latitude and longitude take precedence over ll and ur
    if lat and lng:
        lats, lngs = great_circle_dist(lat, lng, distance / 2.0)

        ll = { "longitude": min(*lngs), "latitude": min(*lats) }
        ur = { "longitude": max(*lngs), "latitude": max(*lats) }

    if ll and ur:
        payload["sceneFilter"]["spatialFilter"] = {
            "filterType": "mbr",
            "lowerLeft": ll,
            "upperRight": ur
        }

    if where:
        payload["sceneFilter"]["metadataFilter"] = {
            "filterType": "value",
            "filterId": where["filter_id"],
            "value": where["value"],
            "operand": "="
        }

    if scene_filter is not None:
        payload["sceneFilter"].update(scene_filter)

    return json.dumps(payload)



import unittest
from unittest import mock

import requests

from usgs import api, USGSError
from .MockPost import MockPost


def check_root_keys(response):

    expected_keys = ["requestId", "version", "sessionId", "data", "errorCode", "errorMessage"]
    for key in expected_keys:
        if key not in response:
            return False

    return True

@mock.patch('usgs.api.requests.Session.post', MockPost('dataset-fields.json'))
def test_dataset_filters():

    expected_keys = ["id", "legacyFieldId", "dictionaryLink", "fieldConfig", "fieldLabel", "searchSql"]
    response = api.dataset_filters("LANDSAT_8_C1")

    assert check_root_keys(response)

    for item in response['data']:
        for key in expected_keys:
            assert key in item

@mock.patch('usgs.api.requests.Session.post', MockPost('download-options.json'))
def test_download_options():

    expected_keys = ["id", "displayId", "entityId", "datasetId", "available", "filesize", "productName", "productCode", "bulkAvailable", "downloadSystem", "secondaryDownloads"]
    response = api.download_options("LANDSAT_8_C1", ["LC82260782020217LGN00"])

    assert check_root_keys(response)

    for item in response["data"]:
        for key in expected_keys:
            assert key in item

@mock.patch('usgs.api.requests.Session.post', MockPost('dataset-download-options.json'))
def test_dataset_download_options():
    expected_keys = ['productId', 'productCode', 'productName']
    response = api.dataset_download_options("LANDSAT_8_C1")

    assert check_root_keys(response)

    for item in response["data"]:
        for key in expected_keys:
            assert key in item

@mock.patch('usgs.api.requests.Session.post', MockPost('download-request.json'))
def test_download_request():

    expected_keys = ['availableDownloads', 'duplicateProducts', 'preparingDownloads', 'failed', 'newRecords', 'numInvalidScenes']
    response = api.download_request("LANDSAT_8_C1", "LC82260782020217LGN00", "5e83d0b84df8d8c2")

    assert check_root_keys

    for key in response["data"]:
        assert key in expected_keys

@mock.patch('usgs.api.requests.Session.post', MockPost('dataset-search.json'))
def test_dataset_search():

    expected_keys = [
        'abstractText', 'acquisitionStart', 'acquisitionEnd', 'catalogs', 'collectionName',
        'collectionLongName', 'datasetId', 'datasetAlias', 'datasetCategoryName',
        'dataOwner', 'dateUpdated', 'doiNumber', 'ingestFrequency', 'keywords',
        'legacyId', 'sceneCount', 'spatialBounds', 'temporalCoverage',
        'supportCloudCover', 'supportDeletionSearch']
    response = api.dataset_search()

    assert check_root_keys(response)
    assert len(response['data']) == 1163
    for key in expected_keys:
        assert key in response['data'][0]

@mock.patch('usgs.api.requests.Session.post', MockPost('scene-metadata.json'))
def test_scene_metadata():

    expected_keys = [
        'browse', 'cloudCover', 'entityId', 'displayId', 'orderingId', 'metadata',
        'hasCustomizedMetadata', 'options', 'selected', 'spatialBounds', 'spatialCoverage',
        'temporalCoverage', 'publishDate']

    response = api.scene_metadata(
        "LANDSAT_8_C1", "LC82260782020217LGN00")
    assert check_root_keys(response)

    for key in expected_keys:
        assert key in response['data']

@mock.patch('usgs.api.requests.Session.post', MockPost('scene-search.json'))
def test_scene_search():
    expected_keys = [
        'browse', 'cloudCover', 'entityId', 'displayId', 'orderingId', 'metadata',
        'hasCustomizedMetadata', 'options', 'selected', 'spatialBounds', 'spatialCoverage',
        'temporalCoverage', 'publishDate']

    where = {
        "filter_id": "5e83d0b8fb079b8b",
        "value": "T1"
    }
    response = api.scene_search(
        "LANDSAT_8_C1", start_date='2016-09-11', end_date='2016-09-12', metadata_type="summary",
        max_results=1, lat=-33.463623929871886, lng=-70.54248332977295, where=where)

    assert check_root_keys(response)
    assert len(response['data']["results"]) == 1

    for expected_key in expected_keys:
        assert expected_key in response['data']["results"][0]


class HttpErrorTests(unittest.TestCase):

    @mock.patch('usgs.api.requests.Session.post', MockPost(status_code=500, reason="Server Error"))
    def test_dataset_filters_http_error(self):
        with self.assertRaises(USGSError) as cm:
            api.dataset_filters("LANDSAT_8_C1")
        assert "HTTP error occurred during dataset filters request" in str(cm.exception)
        assert isinstance(cm.exception.__cause__, requests.HTTPError)

    @mock.patch('usgs.api.requests.Session.post', MockPost(status_code=500, reason="Server Error"))
    def test_download_options_http_error(self):
        with self.assertRaises(USGSError) as cm:
            api.download_options("LANDSAT_8_C1", ["LC82260782020217LGN00"])
        assert "HTTP error occurred during download options request" in str(cm.exception)
        assert isinstance(cm.exception.__cause__, requests.HTTPError)

    @mock.patch('usgs.api.requests.Session.post', MockPost(status_code=500, reason="Server Error"))
    def test_dataset_download_options_http_error(self):
        with self.assertRaises(USGSError) as cm:
            api.dataset_download_options("LANDSAT_8_C1")
        assert "HTTP error occurred during dataset download options request" in str(cm.exception)
        assert isinstance(cm.exception.__cause__, requests.HTTPError)

    @mock.patch('usgs.api.requests.Session.post', MockPost(status_code=500, reason="Server Error"))
    def test_download_request_http_error(self):
        with self.assertRaises(USGSError) as cm:
            api.download_request(
                "LANDSAT_8_C1", "LC82260782020217LGN00", "5e83d0b84df8d8c2")
        assert "HTTP error occurred during download request" in str(cm.exception)
        assert isinstance(cm.exception.__cause__, requests.HTTPError)

    @mock.patch('usgs.api.requests.Session.post', MockPost(status_code=500, reason="Server Error"))
    def test_dataset_search_http_error(self):
        with self.assertRaises(USGSError) as cm:
            api.dataset_search()
        assert "HTTP error occurred during dataset search request" in str(cm.exception)
        assert isinstance(cm.exception.__cause__, requests.HTTPError)

    @mock.patch('usgs.api.requests.Session.post', MockPost(status_code=500, reason="Server Error"))
    def test_login_http_error(self):
        with self.assertRaises(USGSError) as cm:
            api.login("user", "token", save=False)
        assert "HTTP error occurred during login request" in str(cm.exception)
        assert isinstance(cm.exception.__cause__, requests.HTTPError)

    @mock.patch('usgs.api.requests.Session.post', MockPost(status_code=500, reason="Server Error"))
    def test_scene_metadata_http_error(self):
        with self.assertRaises(USGSError) as cm:
            api.scene_metadata("LANDSAT_8_C1", "LC82260782020217LGN00")
        assert "HTTP error occurred during scene metadata request" in str(cm.exception)
        assert isinstance(cm.exception.__cause__, requests.HTTPError)

    @mock.patch('usgs.api.requests.Session.post', MockPost(status_code=500, reason="Server Error"))
    def test_scene_search_http_error(self):
        with self.assertRaises(USGSError) as cm:
            api.scene_search("LANDSAT_8_C1")
        assert "HTTP error occurred during scene search request" in str(cm.exception)
        assert isinstance(cm.exception.__cause__, requests.HTTPError)

    @mock.patch('usgs.api.requests.Session.post', MockPost(status_code=500, reason="Server Error"))
    @mock.patch('usgs.api.os.remove')
    @mock.patch('usgs.api.os.path.exists', return_value=True)
    def test_logout_http_error(self, mock_exists, mock_remove):
        with self.assertRaises(USGSError) as cm:
            api.logout()
        assert "HTTP error occurred during logout request" in str(cm.exception)
        assert isinstance(cm.exception.__cause__, requests.HTTPError)
        # credentials file is always removed on logout, even when the request fails
        mock_remove.assert_called_once()




import os
import json
from usgs import api

testdir = os.path.dirname(os.path.abspath(__file__))


def write_response(response, filename):
    fpath = os.path.join(testdir, 'data', filename)
    with open(fpath, 'w') as f:
        json.dump(response, f)


def create_snapshots():
    """
    Run requests against USGS API for use in tests.
    """

    api_key = api.login(os.environ['USGS_USERNAME'], os.environ['USGS_TOKEN'])

    # Dataset Fields
    response = api.dataset_fields("LANDSAT_8_C1", "EE", api_key=api_key)
    write_response(response, 'dataset-fields.json')

    # Datasets
    response = api.datasets(None, "EE")
    write_response(response, 'datasets.json')

    # Download
    response = api.download("LANDSAT_8_C1", "EE", ["LC80810712017104LGN00"], product='STANDARD')
    write_response(response, 'download.json')

    # Download Options
    response = api.download_options("LANDSAT_8_C1", "EE", ["LC80810712017104LGN00"])
    write_response(response, 'download-options.json')

    # Metadata
    response = api.metadata("LANDSAT_8_C1", "EE", ["LC80810712017104LGN00"])
    write_response(response, 'metadata.json')

    # Search
    response = api.search("LANDSAT_8_C1", "EE", start_date='20170401', end_date='20170402', max_results=10)
    write_response(response, 'search.json')

    api.logout(api_key)


if __name__ == '__main__':
    create_snapshots()




import os
import json

import requests

testdir = os.path.dirname(os.path.abspath(__file__))


class MockPost(object):


    def __init__(self, filename=None, status_code=200, reason="OK"):
        self.filename = filename
        self.status_code = status_code
        self.reason = reason


    def __call__(self, url, payload=None):
        return self


    def raise_for_status(self):
        if self.status_code >= 400:
            raise requests.HTTPError(
                "%d %s" % (self.status_code, self.reason))
        return None


    def json(self):
        filepath = os.path.join(testdir, 'data', self.filename)

        with open(filepath) as f:
            data = json.load(f)

        return data



import json
import pytest
import unittest

from usgs import payloads


def compare_json(s1, s2):
    return json.loads(s1) == json.loads(s2)


class PayloadsTest(unittest.TestCase):

    def test_dataset_filters(self):
        expected = '{"datasetName": "LANDSAT_8_C1"}'
        payload = payloads.dataset_filters("LANDSAT_8_C1")
        assert compare_json(payload, expected)


    def test_download_options(self):
        expected = """{"datasetName": "LANDSAT_8_C1", "entityIds": ["LC80130292014100LGN00", "LC80130282014100LGN00"]}"""
        payload = payloads.download_options("LANDSAT_8_C1", ["LC80130292014100LGN00", "LC80130282014100LGN00"])
        assert compare_json(payload, expected)


    def test_dataset_download_options(self):
        expected = """{"datasetName": "LANDSAT_8_C1"}"""
        payload = payloads.dataset_download_options("LANDSAT_8_C1")
        assert compare_json(payload, expected)


    def test_download_request(self):
        expected = """{"downloads": [{"entityId": "LC80130292014100LGN00", "productId": "5e83d0b84df8d8c2"}], "downloadApplication": "EE"}"""
        payload = payloads.download_request("LANDSAT_8_C1", "LC80130292014100LGN00", "5e83d0b84df8d8c2")
        assert compare_json(payload, expected)


    def test_dataset_search(self):

        expected = """{"datasetName": "LANDSAT_8_C1", "catalog": "EE", "temporalFilter": {"start": "2014-10-01T00:00:00Z", "end": "2014-10-01T23:59:59Z"}, "spatialFilter": {"filterType": "mbr", "lowerLeft": {"latitude": 44.60847, "longitude": -99.69639}, "upperRight": {"latitude": 44.60847, "longitude": -99.69639}}}"""

        ll = {"longitude": -99.69639, "latitude": 44.60847}
        ur = {"longitude": -99.69639, "latitude": 44.60847}
        start_date = "2014-10-01T00:00:00Z"
        end_date = "2014-10-01T23:59:59Z"

        payload = payloads.dataset_search(
            "LANDSAT_8_C1", "EE",
            ll=ll, ur=ur,
            start_date=start_date, end_date=end_date)

        assert compare_json(payload, expected)

    
    def test_login(self):
        expected = """{"username": "username", "token": "token"}"""
        payload = payloads.login("username", "token")
        assert compare_json(payload, expected)


    def test_scene_metadata(self):
        expected = """{"datasetName": "LANDSAT_8_C1", "entityId": "LC82260782020217LGN00", "metadataType": "full"}"""
        payload = payloads.scene_metadata("LANDSAT_8_C1", "LC82260782020217LGN00")
        assert compare_json(payload, expected)


    def test_scene_search(self):
        expected = """{"datasetName": "GLS2005", "maxResults": 3, "startingNumber": null, "metadataType": null, "sceneFilter": {"acquisitionFilter": {"start": "2006-01-01T00:00:00Z", "end": "2007-12-01T00:00:00Z"}, "spatialFilter": {"filterType": "mbr", "lowerLeft": {"longitude": -135, "latitude": 75}, "upperRight": {"longitude": -120, "latitude": 90}}}}"""

        ll = {"longitude": -135, "latitude": 75}
        ur = {"longitude": -120, "latitude": 90}
        start_date = "2006-01-01T00:00:00Z"
        end_date = "2007-12-01T00:00:00Z"

        payload = payloads.scene_search(
            "GLS2005", ll=ll, ur=ur, start_date=start_date,
            end_date=end_date, max_results=3)

        assert compare_json(payload, expected)


    def test_scene_search_radial_distance(self):

        expected = """{"datasetName": "GLS2005", "maxResults": 3, "startingNumber": null, "metadataType": null, "sceneFilter": {"acquisitionFilter": {"start": "2006-01-01T00:00:00Z", "end": "2007-12-01T00:00:00Z"}, "spatialFilter": {"filterType": "mbr", "lowerLeft": {"longitude": -125.01823502237109, "latitude": 84.99840995696343}, "upperRight": {"longitude": -124.98175340746137, "latitude": 85.00158953883317}}}}"""

        lat = 85
        lng = -125
        dist = 1000
        start_date = "2006-01-01T00:00:00Z"
        end_date = "2007-12-01T00:00:00Z"

        payload = payloads.scene_search(
            "GLS2005", lat=lat, lng=lng, distance=dist, start_date=start_date, end_date=end_date, max_results=3)

        assert compare_json(payload, expected), "wrong result: {r} \n for expected: {e}".format(r=payload, e=expected)


    def test_scene_search_with_scene_filter(self):
        expected = """{"datasetName": "GLS2005", "maxResults": 3, "startingNumber": null, "metadataType": null, "sceneFilter": {"acquisitionFilter": {"start": "2006-01-01T00:00:00Z", "end": "2007-12-01T00:00:00Z"}, "ingestFilter": {"start": "2010-01-01", "end": "2012-01-31"}}}"""

        start_date = "2006-01-01T00:00:00Z"
        end_date = "2007-12-01T00:00:00Z"
        scene_filter = {"ingestFilter": {"start": "2010-01-01", "end": "2012-01-31"}}

        payload = payloads.scene_search(
            "GLS2005", start_date=start_date, end_date=end_date,
            max_results=3, scene_filter=scene_filter)

        assert compare_json(payload, expected), "wrong result: {r} \n for expected: {e}".format(r=payload, e=expected)


    def test_scene_search_scene_filter_overrides(self):
        # scene_filter should override values set by other parameters
        expected = """{"datasetName": "GLS2005", "maxResults": 3, "startingNumber": null, "metadataType": null, "sceneFilter": {"acquisitionFilter": {"start": "2010-01-01", "end": "2010-12-31"}}}"""

        start_date = "2006-01-01T00:00:00Z"
        end_date = "2007-12-01T00:00:00Z"
        scene_filter = {"acquisitionFilter": {"start": "2010-01-01", "end": "2010-12-31"}}

        payload = payloads.scene_search(
            "GLS2005", start_date=start_date, end_date=end_date,
            max_results=3, scene_filter=scene_filter)

        assert compare_json(payload, expected), "wrong result: {r} \n for expected: {e}".format(r=payload, e=expected)



import os
import json
from datetime import datetime
from urllib3.util.retry import Retry

import requests
from requests.adapters import HTTPAdapter
from requests_futures.sessions import FuturesSession

from usgs import USGS_API, USGSError, USGSAuthExpiredError, __version__
from usgs import payloads


TMPFILE = os.path.join(os.path.expanduser("~"), ".usgs")


def _get_api_key(api_key):

    if api_key is None and os.path.exists(TMPFILE):
        with open(TMPFILE, "r") as f:
            api_key_info = json.load(f)
        api_key = api_key_info['apiKey']

    return api_key

def _check_for_requests_error(response, message_context=""):
    try:
        response.raise_for_status()
    except requests.HTTPError as e:
        msg = f"HTTP error occurred during {message_context} request, {e}"
        raise USGSError(msg) from e

def _check_for_usgs_error(data):

    error_code = data['errorCode']
    if error_code is None:
        return

    error = data['errorMessage']

    if error_code == 'AUTH_EXPIRED':
        raise USGSAuthExpiredError("API key has expired. Try logging out and logging back in.")

    raise USGSError('%s: %s' % (error_code, error))

def _create_session(api_key):
    api_key = _get_api_key(api_key)

    headers = {
        'User-Agent': 'Python usgs v{}'.format(__version__)
    }
    if api_key:
        headers['X-Auth-Token'] = api_key

    session = requests.Session()
    session.headers.update(headers)
    retries = Retry(total=5, backoff_factor=2)
    session.mount(USGS_API, HTTPAdapter(max_retries=retries))

    return session

def dataset_filters(dataset, api_key=None):
    api_key = _get_api_key(api_key)

    with _create_session(api_key) as session:
        url = '{}/dataset-filters'.format(USGS_API)
        payload = payloads.dataset_filters(dataset)
        r = session.post(url, payload)

    _check_for_requests_error(r, "dataset filters")

    response = r.json()

    _check_for_usgs_error(response)

    return response

def download_options(dataset, entity_ids, api_key=None):
    api_key = _get_api_key(api_key)

    with _create_session(api_key) as session:
        url = '{}/download-options'.format(USGS_API)
        payload = payloads.download_options(dataset, entity_ids)
        r = session.post(url, payload)

    _check_for_requests_error(r, "download options")

    response = r.json()

    _check_for_usgs_error(response)

    return response

def dataset_download_options(dataset, api_key=None):
    """
    The dataset download options method is used to discover downloadable
    products for a specified dataset. Unlike the `download_options` method,
    this does not check product availability.

    :param str dataset: Used to identify the which dataset to return results for.
    """
    api_key = _get_api_key(api_key)

    with _create_session(api_key) as session:
        url = '{}/dataset-download-options'.format(USGS_API)
        payload = payloads.dataset_download_options(dataset)
        r = session.post(url, payload)

    _check_for_requests_error(r, "dataset download options")

    response = r.json()

    _check_for_usgs_error(response)

    return response

def download_request(dataset, entity_id, product_id, api_key=None):
    """
    This method is used to insert the requested downloads into the download queue
    and returns the available download URLs.
    """
    api_key = _get_api_key(api_key)

    with _create_session(api_key) as session:
        url = '{}/download-request'.format(USGS_API)
        payload = payloads.download_request(dataset, entity_id, product_id)
        r = session.post(url, payload)

    _check_for_requests_error(r, "download request")

    response = r.json()

    _check_for_usgs_error(response)

    return response

def dataset_search(dataset=None, catalog=None, ll=None, ur=None, start_date=None, end_date=None, api_key=None):
    api_key = _get_api_key(api_key)

    with _create_session(api_key) as session:
        url = '{}/dataset-search'.format(USGS_API)
        payload = payloads.dataset_search(
            dataset=dataset, catalog=catalog, start_date=start_date,
            end_date=end_date, ll=ll, ur=ur)
        r = session.post(url, payload)

    _check_for_requests_error(r, "dataset search")

    response = r.json()

    _check_for_usgs_error(response)

    return response

def login(username, token, save=True):
    """
    Log in, creating a temporary API key and optionally storing it for later use.

    :param str username: Username of the USGS account to log in with.
    :param str token: Application Token of the USGS account to log in with.
    :param bool save: If true, the API key will be stored in a local file (~/.usgs)
        until `api.logout` is called to remove it. The stored key will be used by
        other functions to authenticate requests whenever an API key is not explicitly
        provided.
    """
    url = '{}/login-token'.format(USGS_API)
    payload = payloads.login(username, token)

    with _create_session(api_key=None) as session:
        created = datetime.now().isoformat()
        r = session.post(url, payload)

    _check_for_requests_error(r, "login")

    response = r.json()

    _check_for_usgs_error(response)

    api_key = response["data"]

    if api_key is None:
        raise USGSError(response.get("errorMessage", "Authentication failed"))

    if save:
        with open(TMPFILE, "w") as f:
            json.dump({
                "apiKey": api_key,
                "created": created
            }, f)

    return response

def logout():
    """
    Log out by deactivating and removing the stored API key, if one exists.
    """
    if not os.path.exists(TMPFILE):
        return

    url = '{}/logout'.format(USGS_API)

    try:
        with _create_session(api_key=None) as session:
            r = session.post(url)

        _check_for_requests_error(r, "logout")

        response = r.json()
        try:
            _check_for_usgs_error(response)
        except USGSAuthExpiredError:
            pass

        return response
    finally:
        os.remove(TMPFILE)

def scene_metadata(dataset, entity_id, api_key=None):
    """
    Request metadata for a given scene in a USGS dataset.

    :param str dataset:
    :param str entity_id:
    :param str api_key:
    """
    api_key = _get_api_key(api_key)

    url = '{}/scene-metadata'.format(USGS_API)
    payload = payloads.scene_metadata(dataset, entity_id)

    with _create_session(api_key) as session:
        r = session.post(url, payload)

    _check_for_requests_error(r, "scene metadata")

    response = r.json()

    _check_for_usgs_error(response)
    return response


def scene_search(dataset,
        max_results=5000, metadata_type=None,
        start_date=None, end_date=None,
        ll=None, ur=None,
        lat=None, lng=None, distance=100,
        where=None, scene_filter=None, starting_number=1, sort_order="DESC", api_key=None):
    """
    :param str dataset:
        USGS dataset (e.g. EO1_HYP_PUB, LANDSAT_8)
    :param str lat:
        Latitude
    :param str lng:
        Longitude
    :param int distance:
        Distance in meters used to for a radial search
    :param dict ll:
        Dictionary of longitude/latitude coordinates for the lower left corner
        of a bounding box search. e.g. { "longitude": 0.0, "latitude": 0.0 }
    :param dict ur:
        Dictionary of longitude/latitude coordinates for the upper right corner
        of a bounding box search. e.g. { "longitude": 0.0, "latitude": 0.0 }
    :param str start_date:
        Start date for when a scene has been acquired
    :param str end_date:
        End date for when a scene has been acquired
    :param dict where:
        Dictionary representing key/values for finer grained conditional
        queries. Only a subset of metadata fields are supported. Available
        fields depend on the value of `dataset`, and maybe be found by
        submitting a dataset_filters query.
    :param int max_results:
        Maximum results returned by the server
    :param dict scene_filter:
        Additional filters to apply to the "sceneFilter". Can override filters set using other parameters.
        e.g. {"ingestFilter": {"start": "2010-01-01", "end": "2012-01-31"}}
    :param int starting_number:
        Starting offset for results of a query.
    :param str sort_order:
        Order in which results are sorted. Ascending or descending w.r.t the acquisition date.
    :param str api_key:
        API key for EROS. Required for searching.
    """
    api_key = _get_api_key(api_key)

    with _create_session(api_key) as session:
        url = '{}/scene-search'.format(USGS_API)
        payload = payloads.scene_search(
            dataset, max_results=max_results, metadata_type=metadata_type,
            start_date=start_date, end_date=end_date,
            ll=ll, ur=ur, lat=lat, lng=lng, distance=distance, where=where,
            scene_filter=scene_filter, starting_number=starting_number)
        r = session.post(url, payload)

    _check_for_requests_error(r, "scene search")

    response = r.json()

    _check_for_usgs_error(response)

    return response
